/*
 *  uird.c  v1.1
 *  Universal Infrared Receiver Daemon - program to control stuff with the universal infrared receiver
 *  Copyright (C) 2000, 2001  Carl Bingel, bingel@dacom.se
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <EXTERN.h>
#include <perl.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "uird.h"
#include "defaults.h"

extern void xs_init( void);
extern int init_ir_module( char *tty_name);
extern struct handlers *read_handler_list( char *filename);
extern void parse_cmdline_args( int argc, char **argv, struct options_struct *options);

static PerlInterpreter *my_perl;

void get_irdata_string( int tty, char *data);
void execute_handler( struct handler_list *h, char *ch, int perl_enabled);
void print_help( void);

void code_testing_mode( int tty);
void learning_mode( int tty, struct options_struct *options);
void normal_mode( int tty, struct options_struct *options, char **env);
void perl_mode( int tty, struct options_struct *options, char **env);

void network_mode( int tty, struct options_struct *options, char **env);



/*        			*/
/*		MAIN		*/
/*        			*/
int main( int argc, char **argv, char **env) {
  FILE *fh;
  int tty, k, i;
  struct options_struct *options; 

  // Allocate space and parse command-line options
  if ((options = (struct options_struct *)malloc( sizeof( struct options_struct))) == NULL) {
    fprintf( stderr, "uird: Out of memory.\n");
    exit( -7);
  }
  parse_cmdline_args( argc, argv, options);

  // If daemon mode (-d), fork and write PID-file then exit
  if (options->daemon_mode) {
    if ( (k = fork()) > 0) {
      // Write PID-file
      fh = fopen( PID_FILE, "w");
      if( fh == NULL) {
        fprintf( stderr, "uird: %s: %s\n", PID_FILE, strerror( errno));
        exit( -8);
      }
      fprintf( fh, "%i\n", k);
      fclose( fh);
      exit( 0);
    }
  }


  if ( options->operation_mode == 4) {						/* Print help */
    print_help();
    exit( 0);
  } else if ( options->operation_mode == 5) {					/* Kill running uird daemon */
    fh = fopen( PID_FILE, "r");
    if ( fh == NULL) {
      fprintf( stderr, "uird: Uird does not seem to be running (no PID-file found).\n");
      exit( 0);
    }
    fscanf( fh, "%i", &i);
    fclose( fh);
    if (!kill( i, SIGTERM)) {
      fprintf( stderr, "uird: kill: %s\n", strerror( errno));
    }
    unlink( PID_FILE);
    exit( 0);
  }

  tty = init_ir_module( options->tty_name);

  if ( options->operation_mode == 0)						/*  Normal mode  */
    normal_mode( tty, options, env);
  else if ( options->operation_mode == 1) 					/*  Code-testing-mode  */
    code_testing_mode( tty);
  else if ( options->operation_mode == 2) 					/*  Perl-mode  */
    perl_mode( tty, options, env);
  else if ( options->operation_mode == 3) 					/*  Learing mode */
    learning_mode( tty, options);
  else if ( options->operation_mode == 6) 					/*  Network mode */
    network_mode( tty, options, env);

}


/*                  */
/*  NETWORK MODE    */
/*                  */
void network_mode( int tty, struct options_struct *options, char **env) {
    //unsigned char ch[ 7];
    int sock;
    struct network_message_struct msg;
    struct sockaddr_in dest_addr;
    struct hostent *he;

    // Initialize network-message
    strcpy( msg.magic_id, "UIRD");
    memset( &msg.password, '\0', sizeof( msg.password));
    strcpy( msg.password, options->udp_passwd);
    msg.msgtype=0;

    // Bind socket
    if ( ( he = gethostbyname( options->udp_dst_addr)) == NULL) {  // get the host info
        fprintf( stderr, "uird: error gethostbyname: %s\n", strerror( errno));
        exit( -1);
    }
    if ( ( sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
        fprintf( stderr, "uird: error socket: %s\n", strerror( errno));
        exit( -1);
    }
    dest_addr.sin_family = AF_INET;
    dest_addr.sin_port = htons( options->udp_dst_port);
    dest_addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset( &(dest_addr.sin_zero), '\0', 8); 

    // Send init-packet
    sendto( sock, &msg, sizeof( msg), 0, (struct sockaddr *)&dest_addr, sizeof(struct sockaddr));

    // Clean ch-buffer
    //memset( ch, 0, 7);

    //
    // MAIN LOOP
    //
    while( 1) {
        get_irdata_string(tty, msg.data);
        msg.msgtype = 2;
        sendto( sock, &msg, sizeof( msg), 0, (struct sockaddr *)&dest_addr, sizeof(struct sockaddr));
    }

}

/*				    */
/*	NORMAL MODE		*/
/*				    */
void normal_mode( int tty, struct options_struct *options, char **env) {
  char *parse_argv[] = { "", NULL, NULL };
  unsigned char found, ch[7];   
  struct handlers *p_handlers;
  struct handler_list *hlist;
  int perl_enabled, k, i;

  // Set which perl-script the perl-interpreter should load
  parse_argv[ 1] = options->perl_file_name;

  // Clean ch-buffer
  memset( ch, 0, 7);

  //  Load and initialize PERL if perl-file-name not equals NONE
  if ( (perl_enabled = strcmp( "NONE", options->perl_file_name)) != 0) {
    my_perl = perl_alloc();
    perl_construct( my_perl);
    //  Load and parse PERL-script, exit if error
    if (perl_parse( my_perl, xs_init, 2, parse_argv, env)) 
      exit( -9);
    perl_run( my_perl);
  }

  //  Read handler-list
  p_handlers = read_handler_list( options->config_file_name);

  //
  // MAIN LOOP
  //
  while( 1) {
    get_irdata_string( tty, ch);
    hlist = p_handlers->h_normal;
    found = 0;
    // Loop through list and see if match
    while( hlist != NULL) {
      if ( strncmp( hlist->datastring, ch, 6) == 0) {	  // Match
        execute_handler( hlist, ch, perl_enabled);
        found = 1;
      }
      hlist = hlist->next;
    }
    // If not found and default is specified - execute default
    if ((!found) && (p_handlers->h_default != NULL)) 
      execute_handler( p_handlers->h_default, ch, perl_enabled);
    // Execute "always"-function if specified
    if ( p_handlers->h_always != NULL)
      execute_handler( p_handlers->h_always, ch, perl_enabled);
  }
}

// Execute handler when button has been pressed
void execute_handler( struct handler_list *h, char *ch, int perl_enabled) {
  char *perl_argv[] = { NULL, NULL};
  if ( h->operation == 0) 	// Shell handler
    system( h->command);
  else if (perl_enabled) {	// Perl handler
    perl_argv[ 0] = ch;
    perl_call_argv( h->command, G_DISCARD, perl_argv);
  } else
    fprintf( stderr, "uird: 'PERL_HANDLER %s' specified, but no perl-script loaded.\n", h->command);
}


/*				        */
/*	LEARNING MODE		*/
/*				        */
void learning_mode( int tty, struct options_struct *options) {
  FILE *fh;
  unsigned char ch[7], linebuf[ 150];
  int k;

  printf( "Learning mode, stop with Ctrl-C\n");
  fh = fopen( options->config_file_name, "a");
  if ( fh == NULL) {
    fprintf( stderr, "uird: error opening uird.conf: %s\n", strerror( errno));
    exit( -5);
  }
  while( 1) {
    printf( "\nPlease press the button on the remote you wish to configure.\n");
    get_irdata_string( tty, ch);
    printf( "IR-data: ");
    for( k=0; k<6; k++) {
      printf( "%i%c", (unsigned int)ch[k], (k!=5) ? '-' : '\n' );
      fprintf( fh, "%i%c", (unsigned int)ch[k], (k!=5) ? '-' : ':' );
    }
    printf( "Enter action for that button: ");
    fgets( linebuf, sizeof( linebuf), stdin);
    linebuf[ strlen( linebuf) - 1] = '\0';	// Chop LF
    fprintf( fh, " %s\n", linebuf);
    fflush( fh);
  }    
}


/*			    */
/*	PERL MODE	*/
/*			    */
void perl_mode( int tty, struct options_struct *options, char **env) {
  char *parse_argv[] = { "", NULL, NULL };
  char *perl_argv[] = { NULL, NULL};
  char ch[ 7];

  memset( ch, 0, 7);
  perl_argv[ 0] = ch;

  // Set which perl-script the perl-interpreter should to load
  parse_argv[ 1] = options->perl_file_name;

  /*  Load and initialize PERL  */
  my_perl = perl_alloc();
  perl_construct( my_perl);

  /*  Load and parse PERL-script, exit if error  */
  if (perl_parse( my_perl, xs_init, 2, parse_argv, env))
    exit( -9);
  perl_run( my_perl); 

  printf( "Perl mode\n");
  while( 1) {
    get_irdata_string( tty, ch);
    perl_call_argv( options->perl_handler_name, G_DISCARD, perl_argv);
  }
}


/*				        */
/*	CODE TESTING MODE	*/
/*				        */
void code_testing_mode( int tty) {
  int k;
  unsigned char ch[7];
  printf( "Code testing mode\n");
  while( 1) {
    get_irdata_string( tty, ch);
    printf( "IR-data: ");
    for( k=0; k<6; k++)
      printf( "%i%c", (unsigned int)ch[k], (k!=5) ? '-' : '\n' );
  }
}



// Print help-texts
//
void print_help() {
  fprintf( stderr, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
		"Usage: uird [-N|-C|-L|-P|-h] [-d] [-f <config-file>] [-t <tty>] [-p <perl-file>] [-s <handler-name>]\n",
		"Modes of operation:\n",
		"  -N	Normal mode (default if no mode is specified). Reads \"action-strings\" from config-file and starts\n	listening for ir-data on tty.\n",
		"  -C	Code testing mode. Reads ir-data from tty and displays on screen (for debugging and testing)\n",
		"  -L	Learn mode. Reads ir-data and ask for action to be performed when in normal mode. Outputs captured \"action-strings\"\n",
		"	to config-file\n",
		"  -P	Perl mode. Reads ir-data and call perl-handler for every ir-data received. The ir-data gets sent as a string\n",
		"	of 6 bytes as an argument to the perl-handler\n",
        	"  -U	Network mode (UDP)\n",
		"  -K	Kill running uird in daemon mode\n",
		"  -d	Daemon mode. uird starts and goes to background operation. To be combined with one of the above modes.\n",
		"\nOther parameters\n",
		"  -f <file>	Specifies which config-file to read\n",
		"  -t <tty>	Specifies which tty to use, default /dev/ttyS0\n",
		"  -p <fil>	Specifies which perl-script to be loaded, default uird.pl\n",
		"  -s <handler>	Specifies which perl subroutine to be called everytime a button is pressed when in \"Perl mode\"\n",
        	"  -Uaddr <ipaddr>	When in network-mode, specify destination addr\n",
        	"  -Uport <port>	When in network-mode, specify destination port\n",
        	"  -Upw <password>	When in network-mode, specify password (clear text)\n"
        );
}




/*  Read data from TTY. The program expects the data (6 bytes) to be received within aprox 10ms		*/
/*  or a timeout occurs and the data gets discarded and the function listens for new data.		*/
void get_irdata_string( int tty, char *data) {
  int starttime, received, r;
  char c;
  int done=0;
 
  starttime = received = 0;
  while( !done) {
    r = read( tty, &c, 1);
    if ( r > 0) {
      starttime=1;
      data[ received++] = c;
      if ( received == 6) {		/*  6 bytes received -- done */
        starttime = received = 0;
        done=1;
      }
    } else {
      usleep( 1000);
      if (starttime > 0) {
        starttime++;
        if (starttime > 10)		/*  Time-out receiving characters after aprox. 10ms  (10 * usleep( 1000)..) */
          starttime = received = 0;
      }
    }
  }

}
